@@ -9,10 +9,9 @@ import time
 from pathlib import Path
 from stat import S_IRUSR, S_IWUSR, S_IXUSR, filemode
 
-from platformdirs import user_cache_path, user_config_path
-
 from .exceptions import NFAPINoCredentials
-from .utility import DEFAULT_TOKEN_EXPIRY, EMBED_NET_RESOURCES, ENVIRONMENTS, NET_RESOURCES, RESOURCES, STATUS_CODES, find_generic_resources, get_generic_resource, get_token_cache, http, is_uuidv4, jwt_environment, jwt_expiry, normalize_caseless, plural
+from .utility import (DEFAULT_TOKEN_EXPIRY, EMBED_NET_RESOURCES, ENVIRONMENTS, RESOURCES, STATUS_CODES, find_generic_resources, get_generic_resource_by_url, get_token_cache, get_user_cache_dir, get_user_config_dir, http, is_uuidv4, jwt_environment,
+                      jwt_expiry, normalize_caseless, plural)
 
 
 class Organization:
@@ -44,18 +43,16 @@ class Organization:
                  log_file: str = None,
                  debug: bool = False,
                  logger: logging.Logger = None,
-                 proxy: str = None):
+                 proxy: str = None,
+                 gateway: str = "gateway"):
         """Initialize an instance of organization."""
         # set debug and file if specified and let the calling application dictate logging handlers
         self.log_file = log_file
         self.debug = debug
         if logger:
-            # print(f"using logger '{logger}' from param")
             self.logger = logger
         else:
-            # print("initializing null logger")
             self.logger = logging.getLogger(__name__)
-#        self.logger = logger or logging.getLogger(__name__)
             self.logger.addHandler(logging.NullHandler())
 
         if self.debug:
@@ -68,7 +65,7 @@ class Organization:
             file_handler.setFormatter(formatter)
             self.logger.addHandler(file_handler)
 
-        # verify auth endpoint's server certificate if proxy is type SOCKS or None
+        # do verify Gateway Service's TLS server certificate if proxy is type SOCKS or None
         self.proxy = proxy
         if proxy is None:
             self.proxies = dict()
@@ -83,6 +80,14 @@ class Organization:
             else:
                 self.verify = False
 
+        # users of older versions of nfsupport-cli will send literal None until they upgrade to a version that provides the --gateway option
+        if gateway is None:
+            self.gateway = "gateway"
+        else:
+            self.gateway = gateway
+
+        self.logger.debug(f"got 'gateway' param {self.gateway}")
+
         epoch = round(time.time())
         self.expiry_seconds = 0  # initialize a placeholder for remaining seconds until expiry
         client_id = None
@@ -104,20 +109,20 @@ class Organization:
             'expiry': None,
             'audience': None
         }
-        cache_dir_path = user_cache_path(appname='netfoundry')
+        self.cache_dir_path = get_user_cache_dir()
         token_cache_file_name = self.profile+'.json'
-        config_dir_path = user_config_path(appname='netfoundry')
+        self.config_dir_path = get_user_config_dir()
 
         try:
             # create and correct mode to 0o700
-            cache_dir_path.mkdir(mode=S_IRUSR | S_IWUSR | S_IXUSR, parents=True, exist_ok=True)
-            cache_dir_path.chmod(mode=S_IRUSR | S_IWUSR | S_IXUSR)
+            self.cache_dir_path.mkdir(mode=S_IRUSR | S_IWUSR | S_IXUSR, parents=True, exist_ok=True)
+            self.cache_dir_path.chmod(mode=S_IRUSR | S_IWUSR | S_IXUSR)
         except Exception as e:
-            raise RuntimeError(f"failed to create cache dir '{str(cache_dir_path.resolve())}', caught {e}")
+            raise RuntimeError(f"failed to create cache dir '{str(self.cache_dir_path.resolve())}', caught {e}")
         else:
-            cache_dir_stats = os.stat(cache_dir_path)
+            cache_dir_stats = os.stat(self.cache_dir_path)
             self.logger.debug(f"token cache dir exists with mode {filemode(cache_dir_stats.st_mode)}")
-        self.token_cache_file_path = Path(cache_dir_path / token_cache_file_name)
+        self.token_cache_file_path = Path(self.cache_dir_path / token_cache_file_name)
         self.logger.debug(f"cache file path is computed '{str(self.token_cache_file_path.resolve())}'")
 
         # short circuit if logout only
@@ -142,7 +147,7 @@ class Organization:
             self.logger.debug(f"got token from env NETFOUNDRY_API_TOKEN as {len(self.token)}B")
         else:
             try:
-                token_cache = get_token_cache(self.token_cache_file_path)
+                token_cache = get_token_cache(self)
             except Exception as e:
                 self.token = None
                 self.expiry = None
@@ -157,7 +162,7 @@ class Organization:
         # if the token was found but not the expiry then try to parse to extract the expiry so we can enforce minimum lifespan seconds
         if self.token and not self.expiry:
             try:
-                self.expiry = jwt_expiry(self.token)
+                self.expiry = jwt_expiry(setup=self)
             except Exception as e:
                 self.expiry = round(epoch + DEFAULT_TOKEN_EXPIRY)
                 self.expiry_seconds = DEFAULT_TOKEN_EXPIRY
@@ -215,7 +220,7 @@ path to credentials file.
                     },
                     {
                         "scope": "site",
-                        "path": config_dir_path
+                        "path": self.config_dir_path
                     },
                 ]
                 for scope in default_creds_scopes:
@@ -247,33 +252,28 @@ path to credentials file.
         # the try-except block is to soft-fail all attempts to parse the JWT,
         # which is intended for the API, not this application
         if self.token and not self.environment:
-            try:
-                self.environment = jwt_environment(self.token)
-            except Exception as e:
-                # an exception here is very unlikely because the called
-                # function is designed to provide a sane default in case the
-                # token can't be parsed
-                raise RuntimeError(f"unexpected error extracting environment from JWT, caught {e}")
-            else:
-                self.logger.debug(f"parsed token as JWT and found environment {self.environment}")
-            finally:
-                if self.environment not in ENVIRONMENTS:
-                    self.logger.warn(f"unexpected environment '{self.environment}'")
+            self.environment = jwt_environment(setup=self)
+            self.logger.debug(f"parsed token as JWT and found environment {self.environment}")
+            if self.environment not in ENVIRONMENTS:
+                self.logger.warning(f"unexpected environment '{self.environment}'")
 
         if self.environment and not self.audience:
-            self.audience = f'https://gateway.{self.environment}.netfoundry.io/'
+            self.audience = f'https://{self.gateway}.{self.environment}.netfoundry.io/'
+            self.logger.debug(f"computed audience URL from gateway and environment: {self.audience}")
 
         if self.environment and self.audience:
             if not re.search(self.environment, self.audience):
                 self.logger.error(f"mismatched audience URL '{self.audience}' and environment '{self.environment}'")
                 exit(1)
+            else:
+                self.logger.debug(f"found audience already computed '{self.audience}' and matching environment '{self.environment}'")
 
         # the purpose of this try-except block is to soft-fail all attempts
         # to parse the JWT, which is intended for the API, not this
         # application
         if self.token and not self.expiry:  # if token was obtained in this pass then expiry is already defined by response 'expires_in' property
             try:
-                self.expiry = jwt_expiry(self.token)
+                self.expiry = jwt_expiry(self)
             except Exception as e:
                 raise RuntimeError(f"unexpected error getting expiry from token, caught {e}")
             else:
@@ -296,15 +296,18 @@ path to credentials file.
             # extract the environment name from the authorization URL aka token API endpoint
             if self.environment is None:
                 self.environment = re.sub(r'https://netfoundry-([^-]+)-.*', r'\1', token_endpoint, re.IGNORECASE)
-                self.logger.debug(f"using environment parsed from token_endpoint URL {self.environment}")
+                self.logger.debug(f"using environment parsed from authenticationUrl: {self.environment}")
             # re: scope: we're not using scopes with Cognito, but a non-empty value is required;
             #  hence "/ignore-scope"
-            scope = "https://gateway."+self.environment+".netfoundry.io//ignore-scope"
+            scope = f"https://gateway.{self.environment}.netfoundry.io//ignore-scope"
+            self.logger.debug(f"computed scope URL from 'gateway' and environment: {scope}")
             # we can gather the URL of the API from the first part of the scope string by
             #  dropping the scope suffix
             self.audience = scope.replace(r'/ignore-scope', '')
-            self.logger.debug(f"using audience parsed from token_endpoint URL {self.audience}")
-            # e.g. https://gateway.production.netfoundry.io/
+            self.logger.debug(f"computed audience from authenticationUrl sans the trailing '/ignore-scope': {self.audience}")
+            audience_parts = self.audience.split('.')
+            self.audience = '.'.join([f"https://{self.gateway}"]+audience_parts[1:])
+            self.logger.debug(f"computed audience with substituted param 'gateway': {self.audience}")
             assertion = {
                 "scope": scope,
                 "grant_type": "client_credentials"
@@ -354,7 +357,7 @@ path to credentials file.
                     }
                     self.token_cache_file_path.write_text(json.dumps(token_cache_out, indent=4))
                 except Exception as e:
-                    self.logger.warn(f"failed to cache token in '{str(self.token_cache_file_path.resolve())}', caught {e}")
+                    self.logger.warning(f"failed to cache token in '{str(self.token_cache_file_path.resolve())}', caught {e}")
                 else:
                     self.logger.debug(f"cached token in '{str(self.token_cache_file_path.resolve())}'")
             else:
@@ -421,10 +424,9 @@ path to credentials file.
             self.audience+'identity/v1/api-account-identities/self',
             self.audience+'identity/v1/user-identities/self',
         ]
-        headers = {"authorization": "Bearer " + self.token}
         for url in urls:
             try:
-                caller, status_symbol = get_generic_resource(url=url, headers=headers, proxies=self.proxies, verify=self.verify)
+                caller, status_symbol = get_generic_resource_by_url(setup=self, url=url)
             except Exception as e:
                 self.logger.debug(f"failed to get caller identity from url: '{url}', trying next until last, caught {e}")
             else:
@@ -437,9 +439,8 @@ path to credentials file.
         :param str identity: UUIDv4 of the identity to get
         """
         url = self.audience+'identity/v1/identities/'+identity_id
-        headers = {"authorization": "Bearer " + self.token}
         try:
-            identity, status_symbol = get_generic_resource(url=url, headers=headers, proxies=self.proxies, verify=self.verify)
+            identity, status_symbol = get_generic_resource_by_url(setup=self, url=url)
         except Exception as e:
             raise RuntimeError(f"failed to get identity from url: '{url}', caught {e}")
         else:
@@ -456,7 +457,7 @@ path to credentials file.
             params[param] = kwargs[param]
         for noop in ['sort', 'size', 'page']:
             if params.get(noop):
-                self.logger.warn(f"query param '{noop}' is not supported by Identity Service")
+                self.logger.warning(f"query param '{noop}' is not fully supported by Identity Service")
         if type in ["UserIdentity", "user-identities"]:
             url = self.audience+'identity/v1/user-identities'
         elif type in ["ApiAccountIdentity", "api-account-identities"]:
@@ -465,10 +466,9 @@ path to credentials file.
             url = self.audience+'identity/v1/identities'
         else:
             raise RuntimeError(f"unexpected value for param 'type', got {type}, need one of 'user-identities' or 'api-account-identities'")
-        headers = {"authorization": "Bearer " + self.token}
         try:
             identities = list()
-            for i in find_generic_resources(url=url, headers=headers, proxies=self.proxies, verify=self.verify, **params):
+            for i in find_generic_resources(setup=self, url=url, **params):
                 identities.extend(i)
         except Exception as e:
             raise RuntimeError(f"failed to get identities from url: '{url}', caught {e}")
@@ -493,13 +493,12 @@ path to credentials file.
                 params[k] = v
         for noop in ['size', 'page']:
             if params.get(noop):
-                self.logger.warn(f"query param '{noop}' is not implemented for Authorization Service in this application")
+                self.logger.warning(f"query param '{noop}' is not implemented for Authorization Service in this application")
 
         url = self.audience+'auth/v1/roles'
-        headers = {"authorization": "Bearer " + self.token}
         try:
             roles = list()
-            for i in find_generic_resources(url=url, headers=headers, proxies=self.proxies, verify=self.verify, **params):
+            for i in find_generic_resources(setup=self, url=url, **params):
                 roles.extend(i)
         except Exception as e:
             raise RuntimeError(f"failed to get roles from url: '{url}', caught {e}")
@@ -510,9 +509,8 @@ path to credentials file.
         """Get roles as a collection."""
 
         url = f"{self.audience}auth/v1/roles/{role_id}"
-        headers = {"authorization": "Bearer " + self.token}
         try:
-            role, status_symbol = get_generic_resource(url=url, headers=headers, proxies=self.proxies, verify=self.verify)
+            role, status_symbol = get_generic_resource_by_url(setup=self, url=url)
         except Exception as e:
             raise RuntimeError(f"failed to get role from url: '{url}', caught {e}")
         else:
@@ -528,13 +526,12 @@ path to credentials file.
             params[param] = kwargs[param]
         for noop in ['sort', 'size', 'page']:
             if params.get(noop):
-                self.logger.warn(f"query param '{noop}' is not supported by Identity Service")
+                self.logger.warning(f"query param '{noop}' is not supported by Identity Service")
 
         url = self.audience+'identity/v1/organizations'
-        headers = {"authorization": "Bearer " + self.token}
         try:
             organizations = list()
-            for i in find_generic_resources(url=url, headers=headers, proxies=self.proxies, verify=self.verify, **params):
+            for i in find_generic_resources(setup=self, url=url, **params):
                 organizations.extend(i)
         except Exception as e:
             raise RuntimeError(f"failed to get organizations from url: '{url}', caught {e}")
@@ -549,9 +546,8 @@ path to credentials file.
         :param id: the UUID of the org
         """
         url = self.audience+'identity/v1/organizations/'+id
-        headers = {"authorization": "Bearer " + self.token}
         try:
-            organization, status_symbol = get_generic_resource(url=url, headers=headers, proxies=self.proxies, verify=self.verify)
+            organization, status_symbol = get_generic_resource_by_url(setup=self, url=url)
         except Exception as e:
             raise RuntimeError(f"failed to get organization from url: '{url}', caught {e}")
         else:
@@ -563,10 +559,9 @@ path to credentials file.
 
         :param network_group_id: the UUID of the network group
         """
-        url = self.audience+'rest/v1/network-groups/'+network_group_id
-        headers = {"authorization": "Bearer " + self.token}
+        url = self.audience+'core/v2/network-groups/'+network_group_id
         try:
-            network_group, status_symbol = get_generic_resource(url=url, headers=headers, proxies=self.proxies, verify=self.verify)
+            network_group, status_symbol = get_generic_resource_by_url(setup=self, url=url)
         except Exception as e:
             raise RuntimeError(f"failed to get network_group from url: '{url}', caught {e}")
         else:
@@ -581,8 +576,6 @@ path to credentials file.
                 "create" is useful for comparing an existing entity to a set of properties that are used to create the same type of
                 entity in a POST request, and "update" may be used in the same way for a PUT update.
         """
-        headers = dict()
-        headers["authorization"] = "Bearer " + self.token
         params = dict()
         requested_types = list()
         valid_types = set()
@@ -599,7 +592,7 @@ path to credentials file.
             self.logger.debug(f"requesting embed of: '{valid_types}'")
 
         url = self.audience+'core/v2/networks/'+network_id
-        network, status_symbol = get_generic_resource(url=url, headers=headers, accept=accept, proxies=self.proxies, verify=self.verify, **params)
+        network, status_symbol = get_generic_resource_by_url(setup=self, url=url, accept=accept, **params)
         return(network)
 
     def find_network_groups_by_organization(self, **kwargs):
@@ -607,10 +600,9 @@ path to credentials file.
 
         :param str kwargs: filter results by any supported query param
         """
-        url = self.audience+'rest/v1/network-groups'
-        headers = {"authorization": "Bearer " + self.token}
+        url = self.audience+'core/v2/network-groups'
         network_groups = list()
-        for i in find_generic_resources(url=url, headers=headers, embedded=RESOURCES['network-groups']._embedded, proxies=self.proxies, verify=self.verify, **kwargs):
+        for i in find_generic_resources(setup=self, url=url, embedded=RESOURCES['network-groups']._embedded, **kwargs):
             network_groups.extend(i)
         return(network_groups)
     get_network_groups_by_organization = find_network_groups_by_organization
@@ -625,7 +617,6 @@ path to credentials file.
         :param bool deleted: include resource entities that have a non-null property deletedAt
         """
         url = self.audience+'core/v2/networks'
-        headers = {"authorization": "Bearer " + self.token}
         params = {
             "findByName": name
         }
@@ -635,7 +626,7 @@ path to credentials file.
             params['status'] = 'DELETED'
         try:
             networks = list()
-            for i in find_generic_resources(url=url, headers=headers, embedded=RESOURCES['networks']._embedded, accept=accept, proxies=self.proxies, verify=self.verify, **params):
+            for i in find_generic_resources(setup=self, url=url, embedded=RESOURCES['networks']._embedded, accept=accept, **params):
                 networks.extend(i)
         except Exception as e:
             raise RuntimeError(f"failed to get networks from url: '{url}', caught {e}")
@@ -688,10 +679,9 @@ path to credentials file.
             params['status'] = "DELETED"
 
         url = self.audience+'core/v2/networks'
-        headers = {"authorization": "Bearer " + self.token}
         try:
             networks = list()
-            for i in find_generic_resources(url=url, headers=headers, embedded=RESOURCES['networks']._embedded, accept=accept, proxies=self.proxies, verify=self.verify, **params):
+            for i in find_generic_resources(setup=self, url=url, embedded=RESOURCES['networks']._embedded, accept=accept, **params):
                 networks.extend(i)
         except Exception as e:
             raise RuntimeError(f"failed to get networks from url: '{url}', caught {e}")
